#!/usr/bin/python3

import sys, getopt

class OptionsError(Exception):

    def __init__(self, idx, errstr): # an error message index, string
        self.idx = idx
        self.errstr = errstr
        super().__init__(errstr)

    def __str__(self):
        return('Error index %d: %s' % (self.idx, self.errstr))

    def __repr__(self):
        return self.__str__()

# Param is a class to declare legal parameters, and to hold information
# about parameters after the command line has been parsed.
# instances of this class are used by including them in a dictionary
# passed to the Options constructor.
# the key to each Param in that dictionary is the short form of the option
# the longName parameter here is the long name, e.g.:
#   myfile -g xyz --gather stuff

class Param:

    # shortName - single character short name for the option
    # longName - long name for option
    # required - iterable of requisite companion options (must contain this option)
    # optional - iterable of optional companion options (can contain this option)
    # paramFlag - True if parameter carries a variable, like --option=/path/to/file
    # sanityCheck - function to check sanity of this parameter (run from Options constructor)
    # paramFunc - function to process this param, defaults to None
    # args - non keyword args for paramFunc
    # kwargs - keyword args for paramFunc

    def __init__(self, shortName, longName, required, optional, *args, paramFlag = False, sanityCheck = None, paramFunc = None, **kwargs):
        self.shortName = shortName
        self.longName = longName
        self.sanityCheck = sanityCheck
        self.required = set(required)
        self.optional = set(optional)
        self.paramFlag = paramFlag # set true if an additional param is required. 
        self.parameter = None # set to value of addtional parameter, e.g., 
                              # --output /tmp/junk.out
        self.active = False # set to true if included in command line
        self.paramFunc = paramFunc
        self.args = args
        self.kwargs = kwargs

    # the Options class holds all the Params, and checks sanity for required
    # and optional parameters

class Options:

    def __init__(self, args, opts):

        """
args - command line arguments to parse
opts is an array of params
data is instance of type Param
        """
        self.args = args # command line arguments
        self.opts = {}
        self.xlateLong = {} # key is long name, data is shortname
        for k in opts:
            self.addParam(k)
        so = '' # build short options string
        lo = [] # build long options array
        for k in self.opts:
            so += k # build a string of short names to pass to getopt
            lo.append(self.opts[k].longName) # build array of long names for getopt
            if self.opts[k].paramFlag: # add ':' and '=' if additional parameter needed
                so += ':'
                lo[-1] += '='
        try:
            self.rawOpts, self.parsedArgs = getopt.getopt(args, so, lo) # parse the command line
        except getopt.GetoptError as e:
            # on error, wipe all args
            self.rawOpts = []
            self.parsedArgs = ()
            raise OptionsError(1, 'Error parsing options')
        # set active status on params that were used
        for o in self.rawOpts: # rawOpts are the options from getopt, parsedArgs are the remaining arguments
            if o[0].startswith ('--'): # if it's a long form name
                ao = self.xlateLong[o[0][2:]] # keep the active options as short names
                self.opts[ao].active = True # flag the option as active
                self.opts[ao].parameter = o[1] # save an associated parameter, if there is one
            else:
                ao  = o[0][1:] # short name, strip off a single hyphen
                self.opts[ao].active = True
                self.opts[ao].parameter = o[1]
        self.sanityCheck()
        if not self.activeOpts() and not self.parsedArgs and 'h' in self.opts:
            self.opts['h'].active = True
            self.rawOpts = ['h']

    def addParam(self, p):
        if p.shortName in self.opts:
            raise OptionsError(2, 'Short option name %s already exists' % p.shortName)
        if p.longName in self.xlateLong:
            raise OptionsError(3, 'Long option name %s already exists' % p.longName)
        self.opts[p.shortName] = p
        self.xlateLong[p.longName] = p.shortName

    # inherit and extend Param to provide more sanity checks
    def sanityCheck(self):
        # check consistency
        # for every option (Param object), in standard processing:
        #   all members of required must be in option list
        #   all additional options must be in optional
        for ao in self.activeOpts():
            if self.opts[ao].sanityCheck:
                self.opts[ao].sanityCheck(self.opts[ao])
            else:
                self.checkForReq(self.opts[ao])
                self.checkValidCompanions(self.opts[ao])

    def checkForReq(self, opt):
        # for each option, make sure required companions are present
        gotReq = True
        for checkOpt in opt.required:
            if not checkOpt in self.activeOpts():
                gotReq = False
                break
        if not gotReq: # a required option error
            reqStr = '' # build list of required options for error text
            for checkOpt in opt.required:
                if reqStr: reqStr += '/'
                reqStr += self.opts[checkOpt].longName
            raise OptionsError(4, '%s required for option %s' % (reqStr, opt.longName))    

    def checkValidCompanions(self, opt):
        # for each option, make sure all companion options are in optional or required
        for checkOpt in self.activeOpts(): # raise error if any option in activeopts not represented
            if not ((checkOpt in opt.required) or (checkOpt in opt.optional) \
                     or (checkOpt == opt.shortName)):
                raise OptionsError(5, 'Option %s not compatible with option %s' % \
                                  (self.opts[checkOpt].longName, opt.longName))
        if ('h' in self.activeOpts()) or (not self.activeOpts() and not self.parsedArgs): # -h or no opts
            for k in self.opts: # disable all params
                self.opts[k].active = False
            self.opts['h'].active = True # turn on -h, which must exist
            self.parsedArgs = () # clear the extra args
            self.rawOpts = ['-h'] # mimic what getopt should have had
  
    def printOpts(self): # display the parameters
        print('%s will run with these parameters:' % sys.argv[0])
        for a in self.activeOpts():
            print('  %s/%s%s' % (self.opts[a].shortName, self.opts[a].longName, ': %s' % \
                                 self.opts[a].parameter if self.opts[a].parameter else ''))
        print('Additional parameters:')
        for a in self.parsedArgs:
            print('  %s' % a)

    def activeOpts(self):
        ao = ''
        for o in self.opts:
            if self.opts[o].active:
                ao += o
        return ao
    
    def execParams(self):
        for opt in self.opts:
            if self.opts[opt].active and self.opts[opt].paramFunc:
                self.opts[opt].paramFunc(self.opts[opt], *self.opts[opt].args, **self.opts[opt].kwargs)

    # get an option, prompt for user input if prompt isn't None
    def fetchOpt(self, option, prompt = None, default = None):
        if self.opts[option].active:
            return self.opts[option].parameter
        elif prompt:
            print(prompt,)
            newval = sys.stdin.readline().strip()
            self.opts[option].parameter = newval
            return newval
        return default

if __name__ == '__main__':

    def generic(param, nameStr, extra):
        print(f'param {param.longName}, name {nameStr}, extra {extra}')

    params = (Param('a', 'optiona', 'b', 'cd', 'first option', paramFunc = generic, extra = 111),
              Param('b', 'optionb', 'a', 'cd', 'second option', paramFlag = True, paramFunc = generic, extra = 222),
              Param('c', 'optionc', '', 'abde', 'third option', paramFunc = generic, extra = 333),
              Param('d', 'optiond', '', 'e', 'fourth option', paramFunc = generic, extra = 444),
              Param('e', 'optione', '', 'd', 'fifth option', paramFunc = generic, extra = 555),
              Param('h', 'help', '', '', 'sixth option', paramFunc = generic, extra = 666))
    try:
        o = Options(sys.argv[1:], params)
        print('Running...')
        o.printOpts()
        o.execParams()
    except OptionsError as e:
        print(e.errstr)

